import random
import string
from enum import Enum
from typing import Any, List, Tuple, Union

try:
    import ujson as json
except ImportError:
    import json


class TelegramBotException(Exception):
    pass


class UpdateField(str, Enum):
    MESSAGE = "message"
    EDITED_MESSAGE = "edited_message"
    CHANNEL_POST = "channel_post"
    EDITED_CHANNEL_POST = "edited_channel_post"
    INLINE_QUERY = "inline_query"
    CHOSEN_INLINE_RESULT = "chosen_inline_result"
    CALLBACK_QUERY = "callback_query"
    SHIPPING_QUERY = "shipping_query"
    PRE_CHECKOUT_QUERY = "pre_checkout_query"
    POLL = "poll"
    POLL_ANSWER = "poll_answer"
    MY_CHAT_MEMBER = "my_chat_member"
    CHAT_MEMBER = "chat_member"
    CHAT_JOIN_REQUEST = "chat_join_request"


class MessageField(str, Enum):
    MESSAGE_ID = "message_id"
    FROM_USER = "from_user"
    SENDER_CHAT = "sender_chat"
    DATE = "date"
    CHAT = "chat"
    FORWARD_FROM = "forward_from"
    FORWARD_FROM_CHAT = "forward_from_chat"
    FORWARD_FROM_MESSAGE_ID = "forward_from_message_id"
    FORWARD_SIGNATURE = "forward_signature"
    FORWARD_SENDER_NAME = "forward_sender_name"
    FORWARD_DATE = "forward_date"
    REPLY_TO_MESSAGE = "reply_to_message"
    VIA_BOT = "via_bot"
    EDIT_DATE = "edit_date"
    MEDIA_GROUP_ID = "media_group_id"
    AUTHOR_SIGNATURE = "author_signature"
    TEXT = "text"
    ENTITIES = "entities"
    ANIMATION = "animation"
    AUDIO = "audio"
    DOCUMENT = "document"
    PHOTO = "photo"
    STICKER = "sticker"
    VIDEO = "video"
    VIDEO_NOTE = "video_note"
    VOICE = "voice"
    CAPTION = "caption"
    CAPTION_ENTITIES = "caption_entities"
    CONTACT = "contact"
    DICE = "dice"
    GAME = "game"
    POLL = "poll"
    VENUE = "venue"
    LOCATION = "location"
    NEW_CHAT_MEMBERS = "new_chat_members"
    LEFT_CHAT_MEMBER = "left_chat_member"
    NEW_CHAT_TITLE = "new_chat_title"
    NEW_CHAT_PHOTO = "new_chat_photo"
    DELETE_CHAT_PHOTO = "delete_chat_photo"
    GROUP_CHAT_CREATED = "group_chat_created"
    SUPERGROUP_CHAT_CREATED = "supergroup_chat_created"
    CHANNEL_CHAT_CREATED = "channel_chat_created"
    MESSAGE_AUTO_DELETE_TIMER_CHANGED = "message_auto_delete_timer_changed"
    MIGRATE_TO_CHAT_ID = "migrate_to_chat_id"
    MIGRATE_FROM_CHAT_ID = "migrate_from_chat_id"
    PINNED_MESSAGE = "pinned_message"
    INVOICE = "invoice"
    SUCCESSFUL_PAYMENT = "successful_payment"
    CONNECTED_WEBSITE = "connected_website"
    PASSPORT_DATA = "passport_data"
    PROXIMITY_ALERT_TRIGGERED = "proximity_alert_triggered"
    VOICE_CHAT_SCHEDULED = "voice_chat_scheduled"
    VOICE_CHAT_STARTED = "voice_chat_started"
    VOICE_CHAT_ENDED = "voice_chat_ended"
    VOICE_CHAT_PARTICIPANTS_INVITED = "voice_chat_participants_invited"
    REPLY_MARKUP = "reply_markup"


class ParseMode(str, Enum):
    MARKDOWN_V2 = "MarkdownV2"
    HTML = "html"
    MARKDOWN = "Markdown"


class PollType(str, Enum):
    QUIZ = "quiz"
    REGULAR = "regular"


class Emoji(str, Enum):
    DICE = "🎲"
    DARTS = "🎯"
    BASKETBALL = "🏀"
    FOOTBALL = "⚽"
    BOWLING = "🎳"
    SLOT_MACHINE = "🎰"


class ChatAction(str, Enum):
    UPLOAD_PHOTO = "upload_photo"
    RECORD_VIDEO = "record_video"
    UPLOAD_VIDEO = "upload_video"
    RECORD_VOICE = "record_voice"
    UPLOAD_VOICE = "upload_voice"
    UPLOAD_DOCUMENT = "upload_document"
    FIND_LOCATION = "find_location"
    RECORD_VIDEO_NOTE = "record_video_note"
    UPLOAD_VIDEO_NOTE = "upload_video_note"
    CHOOSE_STICKER = "choose_sticker"


class MaskPoint(str, Enum):
    FOREHEAD = "forehead"
    EYES = "eyes"
    MOUTH = "mouth"
    CHIN = "chin"


class MIMEType(str, Enum):
    IMAGE_JPEG = "image/jpeg"
    IMAGE_GIF = "image/gif"
    VIDEO_MP4 = "video/mp4"
    AUIDO_OGG = "audio/ogg"
    APPLICATION_PDF = "application/pdf"


class InputFile:
    __slots__ = ("file_name", "_file", "mime_type", "_attach_key")

    def __init__(self,
                 file_name: str,
                 file: Union[str, bytes],
                 mime_type: str = None):
        self.file_name = file_name
        assert isinstance(file, (str, bytes)), True
        self._file = file
        self.mime_type = mime_type
        self._attach_key = None

    @property
    def file_data(self):
        if isinstance(self._file, str):
            with open(self._file, "rb") as file_obj:
                return file_obj.read()
        return self._file

    @property
    def file_tuple(self):
        return (self.file_name, self.file_data,
                self.mime_type) if self.mime_type else (self.file_name,
                                                        self.file_data)

    @property
    def attach_key(self):
        if self._attach_key is None:
            self._attach_key = "".join(
                random.choice(string.ascii_letters + string.digits)
                for _ in range(6))
        return self._attach_key

    @property
    def attach_str(self):
        return "attach://{0}".format(self.attach_key)


class TelegramObject(dict):
    def __init__(self, **kwargs):
        if "from" in kwargs:
            kwargs["from_user"] = kwargs.pop("from")
        super().__init__(kwargs)

    @classmethod
    def __parse__(cls, value):
        if value is not None:
            if isinstance(value, (TelegramObject, str, int, float, bool)):
                return value
            if isinstance(value, dict):
                return TelegramObject(**value)
            if isinstance(value, (tuple, list)):
                return [cls.__parse__(_) for _ in value]
        return value

    def __getitem__(self, name: str) -> Any:
        value = self.__parse__(self.get(name, None))
        self[name] = value
        return value

    def __getattr__(self, name: str) -> Any:
        return self[name]

    @property
    def data_(self):
        return self


Message = CallbackQuery = ChosenInlineResult = InlineQuery = File = User = WebhookInfo = PhotoSize = StickerSet = Location = ShippingAddress = OrderInfo = EncryptedPassportElement = EncryptedCredentials = PassportFile = CallbackGame = GameHighScore = VCard = ShippingQuery = PreCheckoutQuery = Poll = PollAnswer = ChatMemberUpdated = ChatJoinRequst = TelegramObject

MessageEntity = TelegramObject


class MentionEntity(MessageEntity):
    def __init__(self):
        super().__init__(type="mention")


class HashTagEntity(MessageEntity):
    def __init__(self):
        super().__init__(type="hashtag")


class CashTagEntity(MessageEntity):
    def __init__(self):
        super().__init__(type="cashtag")


class BotCommandEntity(MessageEntity):
    def __init__(self):
        super().__init__(type="bot_command")


class URLEntity(MessageEntity):
    def __init__(self):
        super().__init__(type="url")


class EmailEntity(MessageEntity):
    def __init__(self):
        super().__init__(type="email")


class PhoneNumberEntity(MessageEntity):
    def __init__(self):
        super().__init__(type="phone_number")


class BoldEntity(MessageEntity):
    def __init__(self):
        super().__init__(type="bold")


class ItalicEntity(MessageEntity):
    def __init__(self):
        super().__init__(type="italic")


class UnderLineEntity(MessageEntity):
    def __init__(self):
        super().__init__(type="underline")


class StrikeThroughEntity(MessageEntity):
    def __init__(self):
        super().__init__(type="strikethrough")


class CodeEntity(MessageEntity):
    def __init__(self):
        super().__init__(type="code")


class PreEntity(MessageEntity):
    def __init__(self, language: str):
        super().__init__(type="pre", language=language)


class TextLinkEntity(MessageEntity):
    def __init__(self, url: str):
        super().__init__(type="text_link", url=url)


class TextMentionEntity(MessageEntity):
    def __init__(self, user: User):
        super().__init__(type="text_mention", user=user)


class JSONSerializedTelegramObject(TelegramObject):
    @property
    def data_(self):
        return json.dumps(self)


BotCommandScope = JSONSerializedTelegramObject


class BotCommandScopeDefault(BotCommandScope):
    def __init__(self):
        super().__init__(type="default")


class BotCommandScopeAllPrivateChats(BotCommandScope):
    def __init__(self):
        super().__init__(type="all_private_chats")


class BotCommandScopeAllGroupChats(BotCommandScope):
    def __init__(self):
        super().__init__(type="all_group_chats")


class BotCommandScopeAllChatAdministrators(BotCommandScope):
    def __init__(self, chat_id):
        super().__init__(type="all_chat_administrators", chat_id=chat_id)


class BotCommandScopeChat(BotCommandScope):
    def __init__(self, chat_id):
        super().__init__(type="chat", chat_id=chat_id)


class BotCommandScopeChatAdministrators(BotCommandScope):
    def __init__(self, chat_id):
        super().__init__(type="chat_administrators", chat_id=chat_id)


class BotCommandScopeChatMember(BotCommandScope):
    def __init__(self, chat_id, user_id):
        super().__init__(type="chat_member", chat_id=chat_id, user_id=user_id)


class InputMedia(JSONSerializedTelegramObject):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.files = []
        media = self.get("media", None)
        if media and isinstance(media, InputFile):
            self.files.append((media.attach_key, media.file_tuple))
            self["media"] = media.attach_str
        thumb = self.get("thumb", None)
        if thumb and isinstance(thumb, InputFile):
            self.files.append((thumb.attach_key, thumb.file_tuple))
            self["thumb"] = thumb.attach_str


class InputMediaPhoto(InputMedia):
    def __init__(self, media: Union[str, InputFile], **kwargs):
        super().__init__(type="photo", media=media, **kwargs)


class InputMediaVideo(InputMedia):
    def __init__(self,
                 media: Union[InputFile, str],
                 thumb: Union[InputFile, str] = None,
                 **kwargs):
        super().__init__(type="video", media=media, thumb=thumb, **kwargs)


class InputMediaAnimation(InputMedia):
    def __init__(self,
                 media: Union[InputFile, str],
                 thumb: Union[InputFile, str] = None,
                 **kwargs):
        super().__init__(type="animation", media=media, thumb=thumb, **kwargs)


class InputMediaAudio(InputMedia):
    def __init__(self,
                 media: Union[InputFile, str],
                 thumb: Union[InputFile, str] = None,
                 **kwargs):
        super().__init__(type="audio", media=media, thumb=thumb, **kwargs)


class InputMediaDocument(InputMedia):
    def __init__(self,
                 media: Union[InputFile, str],
                 thumb: Union[InputFile, str] = None,
                 **kwargs):
        super().__init__(type="document", media=media, thumb=thumb, **kwargs)


class InlineKeyboardButton(TelegramObject):
    def __init__(self, text: str, **kwargs):
        super().__init__(text=text, **kwargs)


class KeyboardButtonPollType(TelegramObject):
    def __init__(self, poll_type: Union[str, PollType]):
        super().__init__(type=poll_type)


class KeyboardButton(TelegramObject):
    def __init__(self, text: str, **kwargs):
        super().__init__(text=text, **kwargs)


ReplyMarkup = JSONSerializedTelegramObject


class InlineKeyboardMarkup(ReplyMarkup):
    def __init__(self, inline_keyboard):
        super().__init__(inline_keyboard=inline_keyboard)


class ReplyKeyboardMarkup(ReplyMarkup):
    def __init__(self, keyboard, **kwargs):
        super().__init__(keyboard=keyboard, **kwargs)


class ReplyKeyboardRemove(ReplyMarkup):
    def __init__(self, remove_keyboard: bool = True, **kwargs):
        super().__init__(remove_keyboard=remove_keyboard, **kwargs)


class ForceReply(ReplyMarkup):
    def __init__(self, force_reply: bool = True, **kwargs):
        super().__init__(force_reply=force_reply, **kwargs)


class UserProfilePhotos(TelegramObject):
    def __init__(self, total_count: int, photos: List[List[PhotoSize]]):
        super().__init__(total_count=total_count, photos=photos)


class LoginUrl(TelegramObject):
    def __init__(self, url: str, **kwargs):
        super().__init__(url=url, **kwargs)


class BotCommand(TelegramObject):
    def __init__(self, command: str, description: str):
        super().__init__(command=command, description=description)


class Animation(TelegramObject):
    def __init__(self, file_id: str, file_unique_id: str, width: int,
                 length: int, duration: int, **kwargs):
        super().__init__(file_id=file_id,
                         file_unique_id=file_unique_id,
                         width=width,
                         length=length,
                         duration=duration,
                         **kwargs)


class MaskPosition(TelegramObject):
    def __init__(self, point: str, x_shift: float, y_shift: float,
                 scale: float, **kwargs):
        super().__init__(point=point,
                         x_shift=x_shift,
                         y_shift=y_shift,
                         scale=scale,
                         **kwargs)

    @property
    def data_(self):
        return json.dumps(self)


class Sticker(TelegramObject):
    def __init__(self, file_id: str, file_unique_id: str, width: int,
                 length: int, **kwargs):
        super().__init__(file_id=file_id,
                         file_unique_id=file_unique_id,
                         width=width,
                         length=length,
                         **kwargs)


class LabeledPrice(TelegramObject):
    def __init__(self, label: str, amount: int, **kwargs):
        super().__init__(label=label, amount=amount, **kwargs)


class ShippingOption(TelegramObject):
    def __init__(self, id: str, title: str, prices: Tuple[LabeledPrice],
                 **kwargs):
        super().__init__(id=id, title=title, prices=prices, **kwargs)


class ChatPermissions(JSONSerializedTelegramObject):
    def __init__(
        self,
        can_send_messages: bool = True,
        can_send_media_messages: bool = True,
        can_send_polls: bool = True,
        can_send_other_messages: bool = True,
        can_add_web_page_previews: bool = True,
        can_change_info: bool = True,
        can_invite_users: bool = True,
        can_pin_messages: bool = True,
    ):
        super().__init__(
            can_send_messages=can_send_messages,
            can_send_media_messages=can_send_media_messages,
            can_send_polls=can_send_polls,
            can_send_other_messages=can_send_other_messages,
            can_add_web_page_previews=can_add_web_page_previews,
            can_change_info=can_change_info,
            can_invite_users=can_invite_users,
            can_pin_messages=can_pin_messages,
        )


class PassportElementType(str, Enum):
    PERSONAL_DETAILS = "personal_details"
    PASSPORT = "passport"
    INTERNAL_PASSPORT = "internal_passport"
    DRIVER_LICENSE = "driver_license"
    IDENTITY_CARD = "identity_card"
    ADDRESS = "address"
    UTILITY_BILL = "utility_bill"
    BANK_STATEMENT = "bank_statement"
    RENTAL_AGREEMENT = "rental_agreement"
    PASSPORT_REGISTRATION = "passport_registration"
    TEMPORARY_REGISTRATION = "temporary_registration"
    PHONE_NUMBER = "phone_number"
    EMAIL = "email"


class PassportElementError(TelegramObject):
    __slots__ = (
        "source",
        "type",
        "message",
        "field_name",
        "data_hash",
        "file_hash",
        "file_hashes",
        "element_hash",
    )

    def __init__(self, source: str, type: Union[str, PassportElementType],
                 **kwargs):
        super().__init__(
            source=source,
            type=type.value if isinstance(type, PassportElementType) else type,
            **kwargs)


class PassportElementErrorDataField(PassportElementError):
    def __init__(
        self,
        type: Union[str, PassportElementType],
        field_name: str,
        data_hash: str,
        message: str,
    ):
        super().__init__("data",
                         type,
                         field_name=field_name,
                         data_hash=data_hash,
                         message=message)


class PassportElementErrorFrontSide(PassportElementError):
    def __init__(self, type: Union[str, PassportElementType], file_hash: str,
                 message: str):
        super().__init__("front_side",
                         type,
                         file_hash=file_hash,
                         message=message)


class PassportElementErrorReverseSide(PassportElementError):
    def __init__(self, type: Union[str, PassportElementType], file_hash: str,
                 message: str):
        super().__init__("reverse_side",
                         type,
                         file_hash=file_hash,
                         message=message)


class PassportElementErrorSelfie(PassportElementError):
    def __init__(self, type: Union[str, PassportElementType], file_hash: str,
                 message: str):
        super().__init__("selfie", type, file_hash=file_hash, message=message)


class PassportElementErrorFile(PassportElementError):
    def __init__(self, type: Union[str, PassportElementType], file_hash: str,
                 message: str):
        super().__init__("file", type, file_hash=file_hash, message=message)


class PassportElementErrorFiles(PassportElementError):
    def __init__(
        self,
        type: Union[str, PassportElementType],
        file_hashes: Union[List[str], Tuple[str]],
        message: str,
    ):
        super().__init__("files",
                         type,
                         file_hashes=file_hashes,
                         message=message)


class PassportElementErrorTranslationFile(PassportElementError):
    def __init__(self, type: Union[str, PassportElementType], file_hash: str,
                 message: str):
        super().__init__("translation_file",
                         type,
                         file_hash=file_hash,
                         message=message)


class PassportElementErrorTranslationFiles(PassportElementError):
    def __init__(
        self,
        type: Union[str, PassportElementType],
        file_hashes: Union[List[str], Tuple[str]],
        message: str,
    ):
        super().__init__("translation_files",
                         type,
                         file_hashes=file_hashes,
                         message=message)


class PassportElementErrorUnspecified(PassportElementError):
    def __init__(self, type: Union[str, PassportElementType],
                 element_hash: str, message: str):
        super().__init__("unspecified",
                         type,
                         element_hash=element_hash,
                         message=message)


InputMessageContent = TelegramObject


class InputTextMessageContent(InputMessageContent):
    def __init__(self, message_text: str, **kwargs):
        super().__init__(message_text=message_text, **kwargs)


class InputLocationMessageContent(InputMessageContent):
    def __init__(self, latitude: float, longitude: float, **kwargs):
        super().__init__(latitude=latitude, longitude=longitude, **kwargs)


class InputVenueMessageContent(InputMessageContent):
    def __init__(self, latitude: float, longitude: float, title: str,
                 address: str, **kwargs):
        super().__init__(latitude=latitude,
                         longitude=longitude,
                         title=title,
                         address=address,
                         **kwargs)


class InputContactMessageContent(InputMessageContent):
    def __init__(self, phone_number: str, first_name: str, **kwargs):
        super().__init__(phone_number=phone_number,
                         first_name=first_name,
                         **kwargs)


class InputInvoiceMessageContent(InputMessageContent):
    def __init__(
        self,
        title: str,
        description: str,
        payload: str,
        provider_token: str,
        currency: str,
        prices: Union[List[LabeledPrice], Tuple[LabeledPrice]],
        **kwargs,
    ):
        super().__init__(title=title,
                         description=description,
                         payload=payload,
                         provider_token=provider_token,
                         currency=currency,
                         prices=prices,
                         **kwargs)


InlineQueryResult = TelegramObject


class InlineQueryResultArticle(InlineQueryResult):
    def __init__(self, id: str, title: str,
                 input_message_content: InputMessageContent, **kwargs):
        super().__init__(type="article",
                         id=id,
                         title=title,
                         input_message_content=input_message_content,
                         **kwargs)


class InlineQueryResultPhoto(InlineQueryResult):
    def __init__(self, id: str, photo_url: str, thumb_url: str, **kwargs):
        super().__init__(type="photo",
                         id=id,
                         photo_url=photo_url,
                         thumb_url=thumb_url,
                         **kwargs)


class InlineQueryResultGif(InlineQueryResult):
    def __init__(self, id: str, gif_url: str, thumb_url: str, **kwargs):
        super().__init__(type="gif",
                         id=id,
                         gif_url=gif_url,
                         thumb_url=thumb_url,
                         **kwargs)


class InlineQueryResultMpeg4Gif(InlineQueryResult):
    def __init__(self, id: str, mpeg4_url: str, **kwargs):
        super().__init__(type="mpeg4_gif",
                         id=id,
                         mpeg4_url=mpeg4_url,
                         **kwargs)


class InlineQueryResultAudio(TelegramObject):
    def __init__(self, id: str, audio_url: str, file: str, **kwargs):
        super().__init__(type="audio",
                         id=id,
                         audio_url=audio_url,
                         file=file,
                         **kwargs)


class InlineQueryResultVoice(TelegramObject):
    def __init__(self, id: str, voice_url: str, title: str, **kwargs):
        super().__init__(type="voice",
                         id=id,
                         voice_url=voice_url,
                         title=title,
                         **kwargs)


class InlineQueryResultDocument(TelegramObject):
    def __init__(self, id: str, title: str, document_url: str, mime_type: str,
                 **kwargs):
        super().__init__(type="document",
                         id=id,
                         title=title,
                         document_url=document_url,
                         mime_type=mime_type,
                         **kwargs)


class InlineQueryResultLocation(TelegramObject):
    def __init__(self, id: str, latitude: float, longitude: float, title: str,
                 **kwargs):
        super().__init__(type="location",
                         id=id,
                         latitude=latitude,
                         longitude=longitude,
                         title=title,
                         **kwargs)


class InlineQueryResultVenue(InlineQueryResult):
    def __init__(self, id: str, latitude: float, longitude: float, title: str,
                 address: str, **kwargs):
        super().__init__(type="venue",
                         id=id,
                         latitude=latitude,
                         longitude=longitude,
                         title=title,
                         address=address,
                         **kwargs)


class InlineQueryResultContact(TelegramObject):
    def __init__(self, id: str, phone_number: str, first_name: str, **kwargs):
        super().__init__(type="contact",
                         id=id,
                         phone_number=phone_number,
                         first_name=first_name,
                         **kwargs)


class InlineQueryResultGame(TelegramObject):
    def __init__(self, id: str, game_short_name: str, **kwargs):
        super().__init__(type="game",
                         id=id,
                         game_short_name=game_short_name,
                         **kwargs)


class InlineQueryResultCachedPhoto(TelegramObject):
    def __init__(self, id: str, photo_file_id: str, **kwargs):
        super().__init__(type="photo",
                         id=id,
                         photo_file_id=photo_file_id,
                         **kwargs)


class InlineQueryResultCachedGif(TelegramObject):
    def __init__(self, id: str, gif_file_id: str, **kwargs):
        super().__init__(type="git", id=id, gif_file_id=gif_file_id, **kwargs)


class InlineQueryResultCachedMpeg4Gif(TelegramObject):
    def __init__(self, id: str, mpeg4_file_id: str, **kwargs):
        super().__init__(type="mpeg4_gif",
                         id=id,
                         mpeg4_file_id=mpeg4_file_id,
                         **kwargs)


class InlineQueryResultCachedSticker(TelegramObject):
    def __init__(self, id: str, sticker_file_id: str, **kwargs):
        super().__init__(type="sticker",
                         id=id,
                         sticker_file_id=sticker_file_id,
                         **kwargs)


class InlineQueryResultCachedDocument(TelegramObject):
    def __init__(self, id: str, title: str, document_file_id: str, **kwargs):
        super().__init__(type="document",
                         id=id,
                         title=title,
                         document_file_id=document_file_id,
                         **kwargs)


class InlineQueryResultCachedVideo(TelegramObject):
    def __init__(self, id: str, title: str, video_file_id: str, **kwargs):
        super().__init__(type="video",
                         id=id,
                         title=title,
                         video_file_id=video_file_id,
                         **kwargs)


class InlineQueryResultCachedVoice(TelegramObject):
    def __init__(self, id: str, title: str, voice_file_id: str, **kwargs):
        super().__init__(type="voice",
                         id=id,
                         title=title,
                         voice_file_id=voice_file_id,
                         **kwargs)


class InlineQueryResultCachedAudio(TelegramObject):
    def __init__(self, id: str, audio_file_id: str, **kwargs):
        super().__init__(type="audio",
                         id=id,
                         audio_file_id=audio_file_id,
                         **kwargs)



from collections import UserList

from telegrambotclient.base import InlineKeyboardMarkup, ReplyKeyboardMarkup


class ReplyKeyboard(UserList):
    def __init__(self, rows=None):
        super().__init__(rows)

    def add_buttons(self, *buttons, col: int = 1):
        for idx in range(0, len(buttons), col):
            self.append(buttons[idx:idx + col])

    def add_rows(self, *rows):
        self.data += rows

    def markup(self, **kwargs):
        return ReplyKeyboardMarkup(keyboard=self.data, **kwargs)

    def __add__(self, keyboard):
        self.data += keyboard.data
        return self


class InlineKeyboard(ReplyKeyboard):
    def markup(self):
        return InlineKeyboardMarkup(inline_keyboard=self.data)

    def where(self, callback_data: str):
        for row_idx, row in enumerate(self.data):
            for col_idx, button in enumerate(row):
                if button.get("callback_data", "") == callback_data:
                    return row_idx, col_idx
        raise ValueError("{} is not found".format(callback_data))

    def replace(self, callback_data: str, new_button) -> bool:
        try:
            row_idx, col_idx = self.where(callback_data)
            self.data[row_idx][col_idx] = new_button
            return True
        except ValueError:
            return False

    def remove(self, callback_data: str) -> bool:
        try:
            row_idx, col_idx = self.where(callback_data)
            row = self.data[row_idx]
            del row[col_idx]
            if not row:
                del self.data[row_idx]
            return True
        except ValueError:
            return False

    def group(self, callback_data: str):
        buttons = []
        for row in self.data:
            for button in row:
                if button.get("callback_data", "").startswith(callback_data):
                    buttons.append(button)
        return tuple(buttons)



try:
    import ujson as json
except ImportError:
    import json

import time
from collections import UserDict
from typing import Any

from telegrambotclient.utils import pretty_format


class TelegramStorage:
    __slots__ = ("_data", )

    def __init__(self):
        self._data = {}

    def get_field(self, key: str, field: str, expires: int):
        session_data = self._data.get(key, {})
        session_data["_expires"] = int(time.time()) + expires
        return session_data.get(field, None)

    def update_fields(self, key: str, mapping, expires: int) -> bool:
        session_data = self._data.get(key, {})
        session_data["_expires"] = int(time.time()) + expires
        session_data.update(mapping)
        self._data[key] = session_data
        return True

    def delete_fields(self, key: str, *fields, expires: int) -> bool:
        session_data = self._data.get(key, {})
        current_time = int(time.time())
        if session_data.get("_expires", 0) < current_time:
            return self.delete_key(key)
        session_data["_expires"] = current_time + expires
        for field in fields:
            if field in session_data:
                del session_data[field]
        self._data[key] = session_data
        return True

    def delete_key(self, key: str) -> bool:
        if key in self._data:
            del self._data[key]
            return True
        return False

    def data(self, key: str, expires: int):
        session_data = self._data.get(key, {})
        current_time = int(time.time())
        if session_data.get("_expires", 0) >= current_time:
            session_data["_expires"] = current_time + expires
            self._data[key] = session_data
        return session_data


class SQLiteStorage(TelegramStorage):
    __slots__ = ("_db_conn", )

    def __init__(self, db_conn):
        db_conn.execute("""
            CREATE TABLE IF NOT EXISTS `t_session` (
                `key`        TEXT NOT NULL UNIQUE,
                `data`       TEXT NOT NULL,
                `expires`    INTEGER NOT NULL,
                PRIMARY KEY(`key`)
                )
            """)
        self._db_conn = db_conn

    def get_field(self, key: str, field: str, expires):
        with self._db_conn:
            cur = self._db_conn.execute(
                "SELECT data, expires from t_session WHERE key=?", (key, ))
            row_data = cur.fetchone()
            if row_data:
                current_time = int(time.time())
                if row_data[1] < current_time:
                    return None
                self._db_conn.execute(
                    "UPDATE t_session SET expires=? WHERE key=?",
                    (current_time + expires, key))
                return json.loads(row_data[0]).get(field, None)
            return None

    def update_fields(self, key: str, mapping, expires: int) -> bool:
        with self._db_conn:
            cur = self._db_conn.execute(
                "SELECT data, expires from t_session WHERE key=?", (key, ))
            row_data = cur.fetchone()
            current_time = int(time.time())
            if row_data:
                data = json.loads(row_data[0])
                if row_data[1] >= current_time:
                    data.update(mapping)
                else:
                    data = mapping
                cur = self._db_conn.execute(
                    "UPDATE t_session SET data=?, expires=? WHERE key=?",
                    (json.dumps(data), current_time + expires, key))
                return cur.rowcount > 0

            cur = self._db_conn.execute(
                """
                INSERT OR REPLACE INTO t_session (
                    key,
                    data,
                    expires
                ) VALUES (?, ?, ?)
                """, (key, json.dumps(mapping), current_time + expires))
            return cur.lastrowid >= 0

    def delete_fields(self, key: str, *fields, expires: int) -> bool:
        with self._db_conn:
            cur = self._db_conn.execute(
                "SELECT data, expires from t_session WHERE key=?", (key, ))
            row_data = cur.fetchone()
            if row_data:
                current_time = int(time.time())
                if row_data[1] < current_time:
                    return False
                data = json.loads(row_data[0])
                for field in fields:
                    if field in data:
                        del data[field]
                cur = self._db_conn.execute(
                    "UPDATE t_session SET data=?, expires=? WHERE key=?",
                    (json.dumps(data), current_time + expires, key))
                return cur.rowcount > 0
            return False

    def delete_key(self, key: str) -> bool:
        with self._db_conn:
            cur = self._db_conn.execute("DELETE from t_session WHERE key=? ",
                                        (key, ))
            return cur.rowcount > 0

    def data(self, key: str, expires: int):
        cur = self._db_conn.execute(
            "SELECT data, expires from t_session WHERE key=?", (key, ))
        row_data = cur.fetchone()
        if row_data:
            current_time = int(time.time())
            if row_data[1] < current_time:
                return {}
            self._db_conn.execute("UPDATE t_session SET expires=? WHERE key=?",
                                  (current_time + expires, key))
            return json.loads(row_data[0])
        return {}

    def __del__(self):
        self._db_conn.close()


class RedisStorage(TelegramStorage):
    __slots__ = ("_redis", )

    def __init__(self, redis_client):
        self._redis = redis_client

    def get_field(self, key: str, field: str, expires: int):
        if self._redis.exists(key) != 1:
            return None
        self._redis.expire(key, expires)
        value = self._redis.hget(key, field)
        return json.loads(value)[0] if value else None

    def delete_fields(self, key: str, *fields, expires: int) -> bool:
        if self._redis.exists(key) != 1:
            return True
        self._redis.expire(key, expires)
        return bool(self._redis.hdel(key, *fields))

    def update_fields(self, key: str, field_mapping, expires: int) -> bool:
        data = self._redis.hgetall(key)
        data.update({
            field: json.dumps((value, ))
            for field, value in field_mapping.items()
        })
        self._redis.expire(key, expires)
        return bool(self._redis.hset(key, mapping=data))

    def delete_key(self, key: str) -> bool:
        return bool(self._redis.delete(key))

    def data(self, key: str, expires: int):
        if self._redis.exists(key) != 1:
            return {}
        self._redis.expire(key, expires)
        return {
            field: json.loads(value)[0]
            for field, value in self._redis.hgetall(key).items()
        }

    def __del__(self):
        self._redis.close()


class MongoDBStorage(TelegramStorage):
    __slots__ = ("_session")

    def __init__(self, collection):
        self._session = collection

    def get_field(self, key: str, field: str, expires: int):
        result = None
        current_time = int(time.time())
        result = self._session.find_one_and_update(
            {
                "_id": key,
                "_expires": {
                    "$gte": current_time
                }
            }, {"$set": {
                "_expires": current_time + expires
            }},
            projection=(field, ))
        if result is None:
            self._session.delete_one({"_id": key})
        return result[field] if result and field in result else None

    def update_fields(self, key: str, mapping, expires: int) -> bool:
        current_time = int(time.time())
        mapping.update({"_expires": current_time + expires, "_id": key})
        result = self._session.update_one(
            {
                "_id": key,
                "_expires": {
                    "$gte": current_time
                }
            }, {"$set": mapping})
        if result.matched_count == 0:
            result = self._session.replace_one({"_id": key},
                                               mapping,
                                               upsert=True)

        return result.modified_count > 0 or result.upserted_id is not None

    def delete_fields(self, key: str, *fields, expires: int) -> bool:
        current_time = int(time.time())
        query = {"_id": key, "_expires": {"$gte": current_time}}
        update = {"_expires": current_time + expires}
        result = self._session.update_one(query, {
            "$unset": {field: None
                       for field in fields},
            "$set": update
        })
        count = result.modified_count
        if count == 0:
            result = self._session.delete_one({"_id": key})
            count = result.deleted_count
        return count > 0

    def delete_key(self, key: str) -> bool:
        result = self._session.delete_one({"_id": key})
        return result.deleted_count > 0

    def data(self, key: str, expires: int):
        current_time = int(time.time())
        return self._session.find_one_and_update(
            {
                "_id": key,
                "_expires": {
                    "$gte": current_time
                }
            }, {"$set": {
                "_expires": current_time + expires
            }}) or {}


class TelegramSession(UserDict):
    __slots__ = ("_storage", "id", "expires")

    def __init__(self,
                 session_id: str,
                 storage: TelegramStorage,
                 expires: int = 1800) -> None:
        super().__init__({})
        self._storage = storage
        self.id = session_id
        self.expires = expires if expires > 0 else 1800

    def __getitem__(self, field: str) -> Any:
        value = self.data.get(field, None)
        if value is None:
            value = self._storage.get_field(self.id, field, self.expires)
            self.data[field] = value
        return value

    def get(self, field: str, default=None) -> Any:
        return self[field] or default

    def delete(self, *fields) -> bool:
        for field in fields:
            if field in self.data:
                super().__delitem__(field)
        return self._storage.delete_fields(self.id,
                                           *fields,
                                           expires=self.expires)

    def __delitem__(self, field):
        self.delete(field)

    def __contains__(self, field: str) -> bool:
        return super().__contains__(field) or bool(self[field])

    def __repr__(self):
        return pretty_format(self.__data__)

    def pop(self, field: str, default=None) -> Any:
        value = self[field] or default
        del self[field]
        return value

    def save(self) -> bool:
        return self._storage.update_fields(self.id,
                                           self.data,
                                           expires=self.expires)

    def clear(self) -> bool:
        self.data = {}
        return self._storage.delete_key(self.id)

    @property
    def __data__(self):
        return self._storage.data(self.id, self.expires)



from collections import UserDict, UserList
from typing import Callable

from telegrambotclient.base import (CallbackQuery, ChatJoinRequst,
                                    ChatMemberUpdated, ChosenInlineResult,
                                    InlineQuery, Message, MessageField, Poll,
                                    PollAnswer, PreCheckoutQuery,
                                    ShippingQuery, TelegramBotException,
                                    TelegramObject, UpdateField)
from telegrambotclient.bot import TelegramBot, logger
from telegrambotclient.handler import (
    CallbackQueryHandler, ChannelPostHandler, ChatJoinRequestHandler,
    ChatMemberHandler, ChosenInlineResultHandler, CommandHandler,
    EditedChannelPostHandler, EditedMessageHandler, ErrorHandler,
    ForceReplyHandler, InlineQueryHandler, MessageHandler, MyChatMemberHandler,
    PollAnswerHandler, PollHandler, PreCheckoutQueryHandler,
    ShippingQueryHandler, UpdateHandler, _MessageHandler)
from telegrambotclient.utils import parse_callback_data, pretty_format


async def call_handler(handler: UpdateHandler, *args, **kwargs):
    return bool(await handler(*args, **kwargs))


class ListRoute(UserList):
    def add_handler(self, handler: UpdateHandler):
        for idx, _handler in enumerate(self):
            if _handler.callback_name == handler.callback_name:
                self[idx] = handler
                return
        self.data.append(handler)
        return self

    async def call_handlers(self, bot: TelegramBot, data: TelegramObject):
        for handler in self:
            if await call_handler(handler, bot, data) is bot.stop_call:
                return bot.stop_call
        return bot.next_call


class ErrorRoute(ListRoute):
    async def call_handlers(self, bot: TelegramBot, data: TelegramObject,
                            error: Exception) -> bool:
        for handler in self:
            if isinstance(error, handler.errors) and await call_handler(
                    handler, bot, data, error) is bot.stop_call:
                return bot.stop_call
        return bot.next_call


class CommandRoute(UserDict):
    def add_handler(self, handler: CommandHandler):
        for cmd_text in handler.cmds:
            self[cmd_text] = handler
        return self

    async def call_handlers(self, bot: TelegramBot, message: Message):
        entity = message.entities[0]
        bot_command = message.text[entity.offset:entity.length]
        cmd_text, *bot_username = tuple(bot_command.split("@"))
        #for /start@one_bot arg1 arg2 ...
        if bot_username and bot_username != bot.user.username:
            return bot.stop_call
        handler = self.get(cmd_text, None)
        if handler is None:
            return bot.stop_call
        bot_command, *args = tuple(message.text.split())
        return await call_handler(handler, bot, message, *args)


class ForceReplyRoute(UserDict):
    def add_handler(self, handler: ForceReplyHandler):
        self[handler.callback_name] = handler
        return self

    async def call_handlers(self, bot: TelegramBot, message: Message):
        chat_id = message.chat.id if message.chat else message.from_user.id if message.from_user else None
        if chat_id is None:
            return bot.stop_call
        reply_to_message = bot.get_force_reply(chat_id)
        if not reply_to_message or message.reply_to_message.message_id != reply_to_message[
                "message_id"]:
            return bot.stop_call

        handler = self.get(reply_to_message["callback"], None)
        if handler is None:
            raise TelegramBotException(
                "{0} is not found as a force reply callback".format(
                    reply_to_message["callback"]))

        return await call_handler(
            handler, bot, message, *reply_to_message["args"]
        ) if "args" in reply_to_message else await call_handler(
            handler, bot, message)


class MessageRoute(UserList):
    def add_handler(self, handler: _MessageHandler):
        if handler.fields:
            # must be inserted before processing all message fields' handlers
            self[0:0] = ((set(
                field.value if isinstance(field, MessageField) else field
                for field in handler.fields), handler), )
        else:
            self.append((set(), handler))
        return self

    async def call_handlers(self, bot: TelegramBot, message: Message):
        incoming_message_fields = set(message.keys())
        for message_fields, handler in self:
            if incoming_message_fields & message_fields == message_fields and await call_handler(
                    handler, bot, message) is bot.stop_call:
                return bot.stop_call

        return bot.next_call


class CallbackQueryRoute(UserDict):
    def add_handler(self, handler: CallbackQueryHandler):
        self[handler.data] = handler
        return self

    async def call_handlers(self, bot: TelegramBot,
                            callback_query: CallbackQuery):
        handler = self.get(callback_query.data, None)
        if handler and await call_handler(handler, bot,
                                          callback_query) is bot.stop_call:
            return bot.stop_call
        if "|" in callback_query.data:
            button_name, args = parse_callback_data(callback_query.data)
            handler = self.get(button_name, None)
            if handler and await call_handler(handler, bot, callback_query, *
                                              args) is bot.stop_call:
                return bot.stop_call

        return bot.next_call


class TelegramRouter:
    __slots__ = ("name", "route_map", "_handler_callers")
    UPDATE_FIELD_VALUES = UpdateField.__members__.values()

    def __init__(self, name):
        self.name = name
        self.route_map = {}
        self._handler_callers = {
            UpdateField.MESSAGE: self.call_message_handlers,
            UpdateField.EDITED_MESSAGE: self.call_edited_message_handlers,
            UpdateField.CALLBACK_QUERY: self.call_callback_query_handlers,
            UpdateField.CHANNEL_POST: self.call_channel_post_handlers,
            UpdateField.EDITED_CHANNEL_POST:
            self.call_edited_channel_post_handlers,
            UpdateField.INLINE_QUERY: self.call_inline_query_handlers,
            UpdateField.CHOSEN_INLINE_RESULT:
            self.call_chosen_inline_result_handlers,
            UpdateField.SHIPPING_QUERY: self.call_shipping_query_handlers,
            UpdateField.PRE_CHECKOUT_QUERY:
            self.call_pre_checkout_query_handlers,
            UpdateField.POLL: self.call_poll_handlers,
            UpdateField.POLL_ANSWER: self.call_poll_answer_handlers,
            UpdateField.MY_CHAT_MEMBER: self.call_my_chat_member_handlers,
            UpdateField.CHAT_MEMBER: self.call_chat_member_handlers
        }

    def register_handlers(self, handlers):
        for handler in handlers:
            self.register_handler(handler)

    def register_handler(self, handler: UpdateHandler):
        assert isinstance(handler, UpdateHandler), True
        update_field = handler.update_field
        logger.info("bind a %s handler: '%s@%s'", update_field,
                    handler.callback_name, self.name)

        route_data = self.route_map.get(update_field, {})
        if isinstance(handler, CallbackQueryHandler):
            self.route_map[update_field] = CallbackQueryRoute(
                route_data).add_handler(handler)
            return self
        if isinstance(handler, CommandHandler):
            self.route_map[update_field] = CommandRoute(
                route_data).add_handler(handler)
            if UpdateField.MESSAGE.value not in self.route_map:
                self.route_map[UpdateField.MESSAGE.value] = []
            return self

        if isinstance(handler, ForceReplyHandler):
            self.route_map[update_field] = ForceReplyRoute(
                route_data).add_handler(handler)
            return self

        route_data = self.route_map.get(update_field, [])
        if isinstance(handler, (MessageHandler, EditedMessageHandler,
                                ChannelPostHandler, EditedChannelPostHandler)):
            self.route_map[update_field] = MessageRoute(
                route_data).add_handler(handler)
            return self

        if isinstance(handler, ErrorHandler):
            self.route_map[update_field] = ErrorRoute(route_data).add_handler(
                handler)
            return self

        # for others update handlers, using listroute
        self.route_map[update_field] = ListRoute(route_data).add_handler(
            handler)
        return self

    def register_error_handler(self, callback: Callable, *errors):
        return self.register_handler(ErrorHandler(callback, *errors))

    def register_command_handler(self, callback: Callable, *cmds):
        return self.register_handler(CommandHandler(callback, *cmds))

    def register_force_reply_handler(self, callback: Callable):
        return self.register_handler(ForceReplyHandler(callback))

    def register_message_handler(self, callback: Callable, *fields):
        return self.register_handler(MessageHandler(callback, *fields))

    def register_edited_message_handler(self, callback: Callable, *fields):
        return self.register_handler(EditedMessageHandler(callback, *fields))

    def register_channel_post_handler(self, callback: Callable, *fields):
        return self.register_handler(ChannelPostHandler(callback, *fields))

    def register_edited_channel_post_handler(self, callback: Callable,
                                             *fields):
        return self.register_handler(
            EditedChannelPostHandler(callback, *fields))

    def register_inline_query_handler(self, callback: Callable):
        return self.register_handler(InlineQueryHandler(callback))

    def register_chosen_inline_result_handler(self, callback: Callable):
        return self.register_handler(ChosenInlineResultHandler(callback))

    def register_callback_query_handler(self,
                                        callback: Callable,
                                        callback_data: str = None,
                                        game_short_name: str = None):
        return self.register_handler(
            CallbackQueryHandler(callback=callback,
                                 callback_data=callback_data,
                                 game_short_name=game_short_name))

    def register_shipping_query_handler(self, callback: Callable):
        return self.register_handler(ShippingQueryHandler(callback))

    def register_pre_checkout_query_handler(self, callback: Callable):
        return self.register_handler(PreCheckoutQueryHandler(callback))

    def register_poll_handler(self, callback: Callable):
        return self.register_handler(PollHandler(callback))

    def register_poll_answer_handler(self, callback: Callable):
        return self.register_handler(PollAnswerHandler(callback))

    def register_my_chat_member_handler(self, callback: Callable):
        return self.register_handler(MyChatMemberHandler(callback))

    def register_chat_member_handler(self, callback: Callable):
        return self.register_handler(ChatMemberHandler(callback))

    def register_chat_join_request_handler(self, callback: Callable):
        return self.register_handler(ChatJoinRequestHandler(callback))

    ###################################################################################
    #
    # register handlers with decorators
    #
    ##################################################################################
    def error_handler(
        self,
        *errors,
    ):
        def decorator(callback):
            self.register_error_handler(callback, errors)
            return callback

        return decorator

    def force_reply_handler(self):
        def decorator(callback):
            self.register_force_reply_handler(callback)
            return callback

        return decorator

    def command_handler(self, *cmds):
        def decorator(callback):
            self.register_command_handler(callback, *cmds)
            return callback

        return decorator

    def message_handler(self, *fields):
        def decorator(callback):
            self.register_message_handler(callback, *fields)
            return callback

        return decorator

    def edited_message_handler(self, *fields):
        def decorator(callback):
            self.register_edited_message_handler(callback, *fields)
            return callback

        return decorator

    def channel_post_handler(self, *fields):
        def decorator(callback):
            self.register_channel_post_handler(callback, *fields)
            return callback

        return decorator

    def edited_channel_post_handler(self, *fields):
        def decorator(callback):
            self.register_edited_channel_post_handler(callback, *fields)
            return callback

        return decorator

    def inline_query_handler(self):
        def decorator(callback):
            self.register_inline_query_handler(callback)
            return callback

        return decorator

    def chosen_inline_result_handler(self):
        def decorator(callback):
            self.register_chosen_inline_result_handler(callback)
            return callback

        return decorator

    def callback_query_handler(self,
                               callback_data: str = None,
                               game_short_name: str = None):
        def decorator(callback):
            self.register_callback_query_handler(callback, callback_data,
                                                 game_short_name)
            return callback

        return decorator

    def shipping_query_handler(self):
        def decorator(callback):
            self.register_shipping_query_handler(callback)
            return callback

        return decorator

    def pre_checkout_query_handler(self):
        def decorator(callback):
            self.register_pre_checkout_query_handler(callback)
            return callback

        return decorator

    def poll_handler(self):
        def decorator(callback):
            self.register_poll_handler(callback)
            return callback

        return decorator

    def poll_answer_handler(self):
        def decorator(callback):
            self.register_poll_answer_handler(callback)
            return callback

        return decorator

    def my_chat_member_handler(self):
        def decorator(callback):
            self.register_my_chat_member_handler(callback)
            return callback

        return decorator

    def chat_member_handler(self):
        def decorator(callback):
            self.register_chat_member_handler(callback)
            return callback

        return decorator

    def chat_join_request_handler(self):
        def decorator(callback):
            self.register_chat_join_request_handler(callback)
            return callback

        return decorator

    ##################################################################################
    #
    # handler callers
    #
    ##################################################################################
    @classmethod
    def __parse_update_field_and_data__(cls, update: TelegramObject):
        for name, value in update.items():
            # telegram bot api confirmed: At most one of the optional parameters can be present in any given update.
            if name in cls.UPDATE_FIELD_VALUES and value:
                return name, TelegramObject(**value)
        raise TelegramBotException("unknown update field: {0}".format(
            pretty_format(update)))

    async def dispatch(self, bot: TelegramBot, update: TelegramObject):
        logger.debug(
            "\n----------------------------- update ----------------------------------\n%s",
            pretty_format(update))
        update_field, data = self.__parse_update_field_and_data__(update)
        route = self.route_map.get(update_field, None)
        if route is not None:
            try:
                await self._handler_callers[update_field](bot, data)
            except Exception as error:
                await ErrorRoute(self.route_map.get("error", [])
                                 ).call_handlers(bot, data, error)
                raise error

    async def call_message_handlers(self, bot: TelegramBot, message: Message):
        if message.entities and message.entities[
                0].type == "bot_command" and await CommandRoute(
                    self.route_map.get("command", {})).call_handlers(
                        bot, message) is bot.stop_call:
            return bot.stop_call

        if "reply_to_message" in message and await ForceReplyRoute(
                self.route_map.get("force_reply", {})).call_handlers(
                    bot, message) is bot.stop_call:
            return bot.stop_call
        route = self.route_map.get(UpdateField.MESSAGE.value, {})
        return await MessageRoute(route).call_handlers(bot, message)

    async def call_edited_message_handlers(self, bot: TelegramBot,
                                           edited_message: Message):
        if edited_message.entities and edited_message.entities[
                0].type == "bot_command" and await CommandRoute(
                    self.route_map.get("command", {})).call_handlers(
                        bot, edited_message) is bot.stop_call:
            return bot.stop_call

        if "reply_to_message" in edited_message and await ForceReplyRoute(
                self.route_map.get("force_reply", {})).call_handlers(
                    bot, edited_message) is bot.stop_call:
            return bot.stop_call

        route = self.route_map.get(UpdateField.EDITED_MESSAGE.value, {})
        return await MessageRoute(route).call_handlers(bot, edited_message)

    async def call_channel_post_handlers(self, bot: TelegramBot,
                                         message: Message):
        return await MessageRoute(
            self.route_map.get(UpdateField.CHANNEL_POST.value,
                               {})).call_handlers(bot, message)

    async def call_edited_channel_post_handlers(self, bot: TelegramBot,
                                                message: Message):
        return await MessageRoute(
            self.route_map.get(UpdateField.EDITED_CHANNEL_POST.value,
                               {})).call_handlers(bot, message)

    async def call_callback_query_handlers(self, bot: TelegramBot,
                                           callback_query: CallbackQuery):
        return await CallbackQueryRoute(
            self.route_map.get(UpdateField.CALLBACK_QUERY.value,
                               {})).call_handlers(bot, callback_query)

    async def call_inline_query_handlers(self, bot: TelegramBot,
                                         inline_query: InlineQuery):
        return await ListRoute(
            self.route_map.get(UpdateField.INLINE_QUERY.value,
                               ())).call_handlers(bot, inline_query)

    async def call_chosen_inline_result_handlers(
            self, bot: TelegramBot, chosen_inline_result: ChosenInlineResult):
        return await ListRoute(
            self.route_map.get(UpdateField.CHOSEN_INLINE_RESULT.value,
                               ())).call_handlers(bot, chosen_inline_result)

    async def call_shipping_query_handlers(self, bot: TelegramBot,
                                           shipping_query: ShippingQuery):
        return await ListRoute(
            self.route_map.get(UpdateField.SHIPPING_QUERY.value,
                               ())).call_handlers(bot, shipping_query)

    async def call_pre_checkout_query_handlers(
            self, bot: TelegramBot, pre_checkout_query: PreCheckoutQuery):
        return await ListRoute(
            self.route_map.get(UpdateField.PRE_CHECKOUT_QUERY.value,
                               ())).call_handlers(bot, pre_checkout_query)

    async def call_poll_handlers(self, bot: TelegramBot, poll: Poll):
        return await ListRoute(self.route_map.get(UpdateField.POLL.value, ())
                               ).call_handlers(bot, poll)

    async def call_poll_answer_handlers(self, bot: TelegramBot,
                                        poll_answer: PollAnswer):
        return await ListRoute(
            self.route_map.get(UpdateField.POLL_ANSWER.value,
                               ())).call_handlers(bot, poll_answer)

    async def call_my_chat_member_handlers(
            self, bot: TelegramBot, my_chat_member_updated: ChatMemberUpdated):
        return await ListRoute(
            self.route_map.get(UpdateField.MY_CHAT_MEMBER.value,
                               ())).call_handlers(bot, my_chat_member_updated)

    async def call_chat_member_handlers(
            self, bot: TelegramBot, chat_member_updated: ChatMemberUpdated):
        return await self.call_my_chat_member_handlers(bot,
                                                       chat_member_updated)

    async def call_chat_join_reqeust_handlers(
            self, bot: TelegramBot, chat_join_request: ChatJoinRequst):
        return await ListRoute(
            self.route_map.get(UpdateField.CHAT_JOIN_REQUEST.value,
                               ())).call_handlers(bot, chat_join_request)

    def __repr__(self):
        return pretty_format(self.route_map)



import pprint
from functools import wraps
from io import StringIO
from typing import Any, List, Tuple, Union

try:
    import ujson as json
except ImportError:
    import json

_pp = pprint.PrettyPrinter(indent=2)


def pretty_print(data: Any):
    _pp.pprint(data)


def pretty_format(data: Any) -> str:
    return _pp.pformat(data)


def regex_match(*patterns):
    def decorate(method):
        @wraps(method)
        def wrapper(bot, message, *args, **kwargs):
            if message.text:
                for pattern in patterns:
                    match_result = pattern.match(message.text)
                    if match_result:
                        return method(bot, message, *args, match_result,
                                      **kwargs)
            return bot.next_call

        return wrapper

    return decorate


def i18n(default_lang_code: str = "en"):
    def decorate(method):
        @wraps(method)
        def wrapper(bot, data, *args, **kwargs):
            def _(text,
                  lang_code=data.from_user.language_code
                  if data.from_user else default_lang_code):
                return bot.get_text(lang_code, text)

            kwargs.update({"_": _})
            return method(bot, data, *args, **kwargs)

        return wrapper

    return decorate


def build_callback_data(button_name: str, *value) -> str:
    data = "{0}|{1}".format(button_name, json.dumps(value))
    assert len(data.encode("utf-8")) <= 64, True
    return data


def parse_callback_data(callback_data: str):
    button_name, value = tuple(callback_data.split("|", maxsplit=1))
    return button_name, tuple(json.loads(value))


def compose_message_entities(text_entities: Union[List, Tuple],
                             sep: str = " "):
    with StringIO() as buffer_:
        entities = []
        for text_entity in text_entities:
            if isinstance(text_entity, str):
                buffer_.write(text_entity)
                buffer_.write(sep)
                continue
            text, entity = tuple(text_entity)
            entity["offset"] = buffer_.tell()
            entities.append(entity)
            if isinstance(text, str):
                buffer_.write(text)
                buffer_.write(sep)
                entity["length"] = len(text)
            elif isinstance(text, (list, tuple)):
                inner_text, inner_entities = compose_message_entities(text)
                offset = buffer_.tell()
                for inner_entity in inner_entities:
                    inner_entity["offset"] += offset
                buffer_.write(inner_text)
                buffer_.write(sep)
                entities += inner_entities
                entity["length"] = len(inner_text)
        return buffer_.getvalue(), tuple(entities)



from telegrambotclient.api import TelegramBotAPI
from telegrambotclient.bot import TelegramBot
from telegrambotclient.router import TelegramRouter
from telegrambotclient.storage import TelegramStorage


class TelegramBotClient:
    __slots__ = ("bots", "routers", "name", "api_callers")

    def __init__(self, name: str = None):
        self.bots = {}
        self.routers = {}
        self.api_callers = {}
        self.name = name or "default"

    def router(self, name: str = None) -> TelegramRouter:
        name = name or "default"
        router = self.routers.get(name, None)
        if router is None:
            self.routers[name] = TelegramRouter(name)
            return self.routers[name]
        return router

    def create_bot(self,
                   token: str,
                   bot_api: TelegramBotAPI = None,
                   storage: TelegramStorage = None,
                   i18n_source=None,
                   session_expires: int = 1800) -> TelegramBot:

        bot_api = self.api_callers.get(
            bot_api.host if bot_api else "https://api.telegram.org", None)
        if bot_api is None:
            bot_api = TelegramBotAPI()
            self.api_callers[bot_api.host] = bot_api
        bot = TelegramBot(token, bot_api, storage, i18n_source,
                          session_expires)
        self.bots[token] = bot
        return bot


# a default client
bot_client = TelegramBotClient()

__all__ = ("bot_client", "TelegramBotClient", "TelegramBotAPI", "TelegramBot")



import asyncio
import logging
import sys
from contextlib import contextmanager
from typing import Callable, Dict

from telegrambotclient.api import TelegramBotAPI
from telegrambotclient.base import File, Message, TelegramObject
from telegrambotclient.storage import TelegramSession, TelegramStorage

logger = logging.getLogger("telegram-bot-client")
formatter = logging.Formatter(
    '%(levelname)s %(asctime)s (%(filename)s:%(lineno)d): "%(message)s"')
console_output_handler = logging.StreamHandler(sys.stderr)
console_output_handler.setFormatter(formatter)
logger.addHandler(console_output_handler)
logger.setLevel(logging.INFO)


class TelegramBot:
    SESSION_ID_FORMAT = "{0}:{1}"
    next_call = True
    stop_call = False

    __slots__ = ("token", "bot_api", "storage", "i18n_source",
                 "session_expires", "user")

    def __init__(self,
                 token: str,
                 bot_api: TelegramBotAPI = None,
                 storage: TelegramStorage = None,
                 i18n_source: Dict = None,
                 session_expires: int = 1800):

        self.token = token
        self.bot_api = bot_api or TelegramBotAPI()
        if storage is None:
            logger.warning(
                "You are using a memory session which should be for testing only."
            )
            storage = TelegramStorage()
        self.storage = storage
        self.i18n_source = i18n_source
        self.session_expires = session_expires
        self.user = self.get_me()

    def get_session(self, user_id: int, expires: int = 0):
        return TelegramSession(
            self.SESSION_ID_FORMAT.format(self.user.id, user_id), self.storage,
            expires or self.session_expires)

    def clear_session(self, user_id: int):
        session = self.get_session(user_id)
        session.clear()

    @contextmanager
    def session(self, user_id: int, expires: int = 0):
        session = self.get_session(user_id, expires or self.session_expires)
        try:
            yield session
        finally:
            session.save()

    def join_force_reply(self,
                         user_id: int,
                         reply_to_message: Message,
                         callback: Callable,
                         *force_reply_args,
                         expires: int = 0):
        force_reply_callback_name = "{0}.{1}".format(callback.__module__,
                                                     callback.__name__)
        session = self.get_session(user_id, expires or self.session_expires)
        if force_reply_args:
            session["_reply_to_message"] = {
                "message_id": reply_to_message.message_id,
                "callback": force_reply_callback_name,
                "args": force_reply_args
            }
        else:
            session["_reply_to_message"] = {
                "message_id": reply_to_message.message_id,
                "callback": force_reply_callback_name
            }
        session.save()

    def update_force_reply(self, user_id, reply_to_message, expires: int = 0):
        session = self.get_session(user_id, expires or self.session_expires)
        if "_reply_to_message" in session:
            session["_reply_to_message"].update(
                {"message_id": reply_to_message.message_id})
            session.save()

    def remove_force_reply(self, user_id, expires: int = 0):
        session = self.get_session(user_id, expires or self.session_expires)
        del session["_reply_to_message"]

    def get_force_reply(self, user_id, expires: int = 0):
        session = self.get_session(user_id, expires or self.session_expires)
        return session.get("_reply_to_message", {})

    def get_text(self, lang_code: str, text: str):
        if self.i18n_source:
            lang_source = self.i18n_source.get(lang_code, None)
            if lang_source:
                if isinstance(lang_source, dict):
                    return lang_source.get(text, text)
                return lang_source.gettext(text)
        return text

    def reply_message(self, message: Message, **kwargs):
        kwargs["reply_to_message_id"] = message.message_id
        return self.send_message(chat_id=message.chat.id, **kwargs)

    def get_file_url(self, file_path: str):
        return "{0}{1}".format(
            self.bot_api.host,
            self.bot_api.FILE_URL.format(self.token, file_path))

    def get_file_bytes(self, file_obj: File):
        return self.bot_api.api_caller.get_bytes(
            self.bot_api.FILE_URL.format(self.token, file_obj.file_path),
            chunk_size=file_obj.file_size or 1024,
        )

    def get_deep_link(self,
                      payload: str,
                      host: str = "https://t.me",
                      startgroup: bool = False):
        return "{0}/{1}?{2}={3}".format(
            host, self.user.username, "startgroup" if startgroup else "start",
            payload)

    def run_polling(self,
                    on_update_callback: Callable,
                    offset: int = 0,
                    limit: int = 100,
                    timeout: int = 10,
                    allowed_updates=None):
        if timeout == 0:
            logger.warning(
                "You are using 0 as timeout in long polling which should be used for testing only."
            )
        while True:
            for raw_update in self.get_updates(
                    offset=offset,
                    limit=limit,
                    timeout=timeout,
                    allowed_updates=allowed_updates):
                update = TelegramObject(**raw_update)
                offset = update.update_id + 1
                asyncio.run(on_update_callback(self, update))

    def __getattr__(self, api_name):
        def api_method(**kwargs):
            return getattr(self.bot_api, api_name)(self.token, **kwargs)

        return api_method



import asyncio
from typing import Callable, Union

from telegrambotclient.base import UpdateField


class UpdateHandler:
    __slots__ = ("callback", "update_field")

    def __init__(self, callback: Callable, update_field: Union[UpdateField,
                                                               str]):
        self.callback = callback
        self.update_field = update_field.value if isinstance(
            update_field, UpdateField) else update_field

    @property
    def callback_name(self):
        return "{0}.{1}".format(self.callback.__module__,
                                self.callback.__name__)

    def __repr__(self) -> str:
        return self.callback_name

    async def __call__(self, *args, **kwargs):
        return await self.callback(
            *args, **kwargs) if asyncio.iscoroutinefunction(
                self.callback) else self.callback(*args, **kwargs)


class ErrorHandler(UpdateHandler):
    __slots__ = ("errors", )

    def __init__(self, callback: Callable, *errors):
        super().__init__(callback=callback, update_field="error")
        self.errors = errors or (Exception, )


class CommandHandler(UpdateHandler):
    __slots__ = ("cmds", )

    def __init__(self, callback: Callable, *cmds):
        super().__init__(callback=callback, update_field="command")
        self.cmds = cmds


class ForceReplyHandler(UpdateHandler):
    def __init__(self, callback: Callable):
        super().__init__(callback=callback, update_field="force_reply")


class _MessageHandler(UpdateHandler):
    __slots__ = ("fields", )

    def __init__(self,
                 callback: Callable,
                 update_field: Union[UpdateField, str] = UpdateField.MESSAGE,
                 *fields):
        super().__init__(callback=callback, update_field=update_field)
        self.fields = fields


class MessageHandler(_MessageHandler):
    def __init__(self, callback: Callable, *fields):
        super().__init__(callback, UpdateField.MESSAGE, *fields)


class EditedMessageHandler(_MessageHandler):
    def __init__(self, callback: Callable, *fields):
        super().__init__(callback, UpdateField.EDITED_MESSAGE, *fields)


class ChannelPostHandler(_MessageHandler):
    def __init__(self, callback: Callable, *fields):
        super().__init__(callback, UpdateField.CHANNEL_POST, *fields)


class EditedChannelPostHandler(_MessageHandler):
    def __init__(self, callback: Callable, *fields):
        super().__init__(callback, UpdateField.EDITED_CHANNEL_POST, *fields)


class CallbackQueryHandler(UpdateHandler):
    __slots__ = ("data", )

    def __init__(self,
                 callback: Callable,
                 callback_data: str = None,
                 game_short_name: str = None):
        super().__init__(callback, UpdateField.CALLBACK_QUERY)
        self.data = callback_data or game_short_name
        assert bool(self.data), True


class InlineQueryHandler(UpdateHandler):
    def __init__(self, callback: Callable):
        super().__init__(callback, UpdateField.INLINE_QUERY)


class ChosenInlineResultHandler(UpdateHandler):
    def __init__(self, callback: Callable):
        super().__init__(callback, UpdateField.CHOSEN_INLINE_RESULT)


class ShippingQueryHandler(UpdateHandler):
    def __init__(self, callback: Callable):
        super().__init__(callback, UpdateField.SHIPPING_QUERY)


class PreCheckoutQueryHandler(UpdateHandler):
    def __init__(self, callback: Callable):
        super().__init__(callback, UpdateField.PRE_CHECKOUT_QUERY)


class PollHandler(UpdateHandler):
    def __init__(self, callback: Callable):
        super().__init__(callback, UpdateField.POLL)


class PollAnswerHandler(UpdateHandler):
    def __init__(self, callback: Callable):
        super().__init__(callback, UpdateField.POLL_ANSWER)


class MyChatMemberHandler(UpdateHandler):
    def __init__(self, callback: Callable):
        super().__init__(callback, UpdateField.MY_CHAT_MEMBER)


class ChatMemberHandler(UpdateHandler):
    def __init__(self, callback: Callable):
        super().__init__(callback, UpdateField.CHAT_MEMBER)


class ChatJoinRequestHandler(UpdateHandler):
    def __init__(self, callback: Callable):
        super().__init__(callback, UpdateField.CHAT_JOIN_REQUEST)



try:
    import ujson as json
except ImportError:
    import json

from io import BytesIO
import urllib3

from telegrambotclient.base import (BotCommandScope, InputFile, InputMedia,
                                    TelegramBotException, TelegramObject)


def exclude_none(**kwargs):
    return {key: value for key, value in kwargs.items() if value is not None}


class TelegramBotAPIException(TelegramBotException):
    __slots__ = ("ok", "error_code", "parameters")

    def __init__(self, **kwargs) -> None:
        super().__init__(kwargs["description"])
        self.ok = kwargs["ok"]
        self.error_code = kwargs["error_code"]
        self.parameters = TelegramObject(
            **kwargs["parameters"]) if "parameters" in kwargs else {}


class TelegramBotAPI:
    API_URL = "/bot{0}/{1}"
    FILE_URL = "/file/bot{0}/{1}"
    __slots__ = ("api_caller", "host")

    def __init__(self,
                 host: str = None,
                 maxsize: int = 10,
                 block: bool = True,
                 **pool_kwargs):
        class _TelegramBotAPICaller:
            __slots__ = ("pool", )

            def __init__(_self, maxsize: int, block: bool,
                         **connection_pool_kwargs):
                connection_pool_kwargs.get("headers", {}).update({
                    "connection":
                    "keep-alive",
                    "user-agent":
                    "telegram-bot-client: A Telegram Bot API Python client",
                })
                # from urllib3.connection.default_socket_options
                from socket import IPPROTO_TCP, SOL_SOCKET, TCP_NODELAY, SO_KEEPALIVE
                connection_pool_kwargs[
                    "socket_options"] = connection_pool_kwargs.get(
                        "socket_options", []) + [
                            (IPPROTO_TCP, TCP_NODELAY, 1),
                            (SOL_SOCKET, SO_KEEPALIVE, 1),
                        ]
                if self.host.startswith("https://"):
                    _self.pool = urllib3.HTTPSConnectionPool(
                        host=self.host[8:],
                        maxsize=maxsize,
                        block=block,
                        **connection_pool_kwargs)
                elif self.host.startswith("http://"):
                    _self.pool = urllib3.HTTPConnectionPool(
                        host=self.host[7:],
                        maxsize=maxsize,
                        block=block,
                        **connection_pool_kwargs)
                else:
                    raise TelegramBotException(
                        "Telegram Bot API only supports https:// and http://")

            @classmethod
            def __format_response__(cls, response):
                if response.status == 500:
                    raise TelegramBotException(response.data)
                json_response = json.loads(response.data.decode("utf-8"))
                if response.status == 200 and json_response["ok"]:
                    result = json_response.get("result", None)
                    if result and isinstance(result, dict):
                        return TelegramObject(**result)
                    return result
                raise TelegramBotAPIException(**json_response)

            def request(_self, api_url: str, data={}, files=()):

                if not files:
                    return _self.__format_response__(
                        _self.pool.request("GET", api_url, fields=data))

                for file in files:
                    data[file[0]] = file[1]
                return _self.__format_response__(
                    _self.pool.request("POST", api_url, fields=data))

            def get_bytes(_self, file_path: str, chunk_size: int) -> bytes:
                response = _self.pool.request("GET",
                                              file_path,
                                              preload_content=False)
                try:
                    if response.status == 200:
                        with BytesIO() as buffer:
                            for chunk in response.stream(chunk_size):
                                buffer.write(chunk)
                            return buffer.getvalue()
                    raise TelegramBotException(response.data)
                finally:
                    response.release_conn()

        self.host = host or "https://api.telegram.org"
        self.api_caller = _TelegramBotAPICaller(maxsize=maxsize,
                                                block=block,
                                                **pool_kwargs)

    @classmethod
    def __prepare_request_params__(cls, **kwargs):
        api_data = exclude_none(**kwargs)
        files = []
        for field in tuple(api_data.keys()):
            value = api_data[field]
            if isinstance(value, (str, int, bool, float)):
                continue
            if isinstance(value, (list, tuple)):
                api_data[field] = json.dumps(value)
                continue
            if isinstance(value, TelegramObject):
                api_data[field] = value.data_
                continue
            if isinstance(value, InputFile):
                if field == "thumb":
                    files.append((value.attach_key, value.file_tuple))
                    api_data["thumb"] = value.attach_str
                else:
                    files.append((field, value.file_tuple))
                    del api_data[field]
        return api_data, files

    def call_api(self, token: str, api_name: str, data, files):
        return self.api_caller.request(
            self.API_URL.format(token,
                                api_name.replace("_", "").lower()), data,
            files)

    def send_media_group(self, token: str, chat_id, media, **kwargs):
        assert 2 <= len(media) <= 10, True
        media_files = []
        media_group = []
        for input_media in media:
            assert isinstance(input_media, InputMedia), True
            media_files.extend(input_media.files)
            media_group.append(input_media.data_)
        api_data, files = self.__prepare_request_params__(chat_id=chat_id,
                                                          media=media_group,
                                                          **kwargs)
        return self.call_api(token,
                             "sendMediaGroup",
                             data=api_data,
                             files=media_files + files)

    def edit_message_media(self,
                           token: str,
                           chat_id=None,
                           message_id: int = None,
                           inline_message_id: int = None,
                           media: InputMedia = None,
                           **kwargs):
        assert isinstance(media, InputMedia), True
        api_data, files = self.__prepare_request_params__(
            chat_id=chat_id,
            message_id=message_id,
            inline_message_id=inline_message_id,
            media=media,
            **kwargs)
        return self.call_api(token,
                             "editMessageMedia",
                             data=api_data,
                             files=media.files + files)

    def get_my_commands(self,
                        token: str,
                        scope: BotCommandScope = None,
                        language_code: str = None):
        return tuple([
            TelegramObject(**raw_command) for raw_command in
            self.getMyCommands(token, scope=scope, language_code=language_code)
        ])

    def __getattr__(self, api_name: str):
        def bot_api_method(token: str, **kwargs):
            api_data, files = self.__prepare_request_params__(**kwargs)
            return self.call_api(token, api_name, data=api_data, files=files)

        return bot_api_method



#!/usr/bin/env python
from setuptools import setup

with open("README.md", 'r') as f:
    long_description = f.read()

setup(
    name='py-telegram-bot-client',
    version="5.5.8",
    description='A Telegram Bot API Python Client',
    long_description_content_type="text/markdown",
    long_description=long_description,
    url='https://github.com/ds828/py-telegram-bot-client',
    author='ds828',
    author_email='songdi19@gmail.com',
    packages=['telegrambotclient'],
    install_requires=['urllib3', 'ujson'],
    python_requires=">=3.5",
)
